#include "Curve.h"

#include "Config.h"

#include <iostream>
#include <fstream>
#include <cmath>

#include <windows.h>
#include <Fl/Gl.h>

using namespace std;

typedef Point::triple triple;

Curve::Curve(void) : controls(), curr(), valid_(false), current_norm(0)
{
}

Curve::~Curve(void)
{
}

void Curve::add_normal() {
	if (current_norm == 0 || normals.size() < 2)
		return;

	int i;
	vector<pair<double, double> >::iterator iter;
	for (i = 0, iter = normals.begin(); i < current_norm; ++i, ++iter) {}

	double length = 0.5 * (iter--->first + iter->first);
	double angle1 = iter++->second;
	double angle2 = iter->second;

	if (abs(angle1 - angle2) > 180)
	{
		if (angle1 < angle2) angle1 += 360;
		else angle1 -= 360;
	}

	double angle = 0.5 * (angle1 + angle2);

	normals.insert(iter, pair<double, double>(length, angle));

}

void Curve::write(string filename)
{
	ofstream file;
	file.open(("./tracks/" + filename + ".crv").c_str());

	file << "control points" << endl;
	file << controls.size() << endl;
	
	for (vector<ControlPoint>::iterator iter = controls.begin(); iter != controls.end(); ++iter)
	{
		iter->write(file);
	}

	file << "normal nodes" << endl;
	file << normals.size() << endl;

	for (vector<pair<double, double> >::iterator iter = normals.begin(); iter != normals.end(); ++iter)
	{
		file << iter->first << " " << iter->second << endl;
	}

	file.close();

}

void Curve::read(std::string filename) {
	ifstream file;
	file.open(("./tracks/" + filename + ".crv").c_str());

	if (!file.good()) {
		cout << "file error" << endl;
		return;
	}

	file.ignore(100, '\n');
	int count;
	file >> count;
	
	vector<ControlPoint> cps;
	for (int i = 0; i < count; i++)
	{
		cps.push_back(ControlPoint::read(file));
	}

	file.ignore(100, '\n');
	file.ignore(100, '\n');
	file >> count;

	vector<pair<double, double> > n;
	for (int i = 0; i < count; i++)
	{
		double s, t;
		file >> s;
		file >> t;
		n.push_back(pair<double, double>(s, t));
	}

	controls = cps;
	normals = n;
	valid_ = false;
	curr = cps.end();
	current_norm = 0;

}

void Curve::set(std::vector<ControlPoint>::iterator iter)
{
	unset();
	curr = iter;
	if (curr != controls.end())
		curr->select();
	valid_ = true;
}

void Curve::unset()
{
	if (valid_)
		curr->deselect();
	valid_ = false;
}

void Curve::addControl(ControlPoint c)
{
	vector<ControlPoint>::iterator iter = controls.end();
	if (valid_)
		iter = curr;

	unset();
	iter = controls.insert(iter, c);
	set(iter);	
}

void Curve::drawControls()
{
	vector<ControlPoint>::iterator iter;
	for (iter = controls.begin(); iter != controls.end(); iter++)
		iter->draw();
}

void Curve::next_norm()
{
	current_norm++;
	current_norm %= normals.size();
}

void Curve::previous_norm()
{
	current_norm--;
	while (current_norm < 0)
		current_norm += normals.size();
}

void Curve::drawNormals()
{
	
	vector<pair<double, double> >::iterator iter;
	int i = 0;
	for (iter = normals.begin(); iter != normals.end(); ++iter, ++i)
	{		
		glPushMatrix();
		Point p = simple_point_at_length(iter->first);
		
		Point::triple pos = p.position();
		Point::triple dir = p.direction();

		dir.normalize();

		triple world_up(0, 0, 1);

		triple out, up = dir.cross(world_up).cross(dir);
		out = dir.rotate(up, 90);

		out = out.rotate(dir, iter->second);
		up = up.rotate(dir, iter->second);

		vector<double> mat = Config::mat(dir, out, up);

		double mat_[16];
		for (int j = 0; j < 16; j++)
			mat_[j] = mat[j];
		
		glTranslatef(pos.x, pos.y, pos.z);
		glMultMatrixd(mat_);		

		glColor3f(0, 0, 0);
		if (!i)
			glColor3f(0.5, 0, 0);
		if (i == current_norm)
			glColor3f(0.8, 0.1, 0.1);
		glScalef(10,10,10);
		glBegin(GL_LINES);
			glVertex3f(0, 0, 0);
			glVertex3f(0, 0, 1);
			glVertex3f(0, 0, 1);
			glVertex3f(0, 0.3, 0.7);
			glVertex3f(0, 0, 1);
			glVertex3f(0, -0.3, 0.7);
		glEnd();	
		glPopMatrix();
		
	}


}

bool Curve::pick(double p[3], double d[3], ControlPoint*& cp)
{
	bool hit = false;
	vector<ControlPoint>::iterator iter;

	vector<ControlPoint*> hits;

	for (iter = controls.begin(); iter != controls.end(); iter++)
	{
		if (iter->pick(p, d))
		{
			hit = true;
			hits.push_back(&(*iter));			
		}
	}

	if (hit == true)
	{
		vector<ControlPoint*>::iterator piter = hits.begin();

		cp = *piter;
		double min = (*piter)->squaredDistance(p);

		for (piter++; piter != hits.end(); piter++)
		{
			double sd = (*piter)->squaredDistance(p);
			if (sd < min)
			{
				cp = *piter;
				min = sd;
			}
		}

		return true;
	}

	return false;

}

bool Curve::select(double p[3], double d[3])
{
	vector<ControlPoint>::iterator iter;
	for (iter = controls.begin(); iter != controls.end(); iter++)
				iter->deselect();

	ControlPoint* cp;

	if (pick(p, d, cp))
	{
		for (iter = controls.begin(); iter != controls.end(); iter++)
			if (&*iter == cp) break;

		set(iter);

		return true;
	}

	unset();

	return false;
}

void Curve::remove_point()
{
	if (!valid_)
		return;
	if (controls.size() < 3)
		return;
	unset();
	controls.erase(curr);	
}

bool Curve::hover(double p[3], double d[3])
{
	vector<ControlPoint>::iterator iter;
	for (iter = controls.begin(); iter != controls.end(); iter++)
				iter->deHover();

	ControlPoint* cp;

	if (pick(p, d, cp))
	{
		cp->hover();
		return true;
	}

	return false;

}

Point Curve::simple_point_at_length(double l)
{
	while (l < 0) l += length_.back();
	while (l > length_.back()) l -= length_.back();
	l = max<double>(l, 0.0001);
	int index = 0;
	while (length_[index] < l)
		index++;

	double length1, length2;
	double alpha = (l - length_[index - 1]) / (length_[index] - length_[index - 1]);

	int a = points_.size();
	Point p1 = points_[index - 1];
	Point p2 = points_[index];

	Point p3 = Point::interpolate(alpha, p2, p1);

	return p3;
}

Point Curve::point_at_length(double l)
{
	while (l < 0) l += length_.back();
	while (l > length_.back()) l -= length_.back();
	typedef Point::triple triple;

	Point p3 = simple_point_at_length(l);

	triple final_pos = p3.position();
	triple final_dir = p3.direction();

	// now calculating norm... this is sloppy.
	int index = 0;

	while (normals[index].first < l && index < normals.size() - 1)
		++index;

	if (normals[index].first < l)
		index = 0;

	pair<double, double> n1, n2;
	
	n2 = normals[index];
	if (index != 0) 
		n1 = normals[index - 1];
	else {
		n1 = normals.back();
		n2 = pair<double, double>(n2);
		n2.first += length();
	}

	double n2_angle = n2.second;

	if (abs(n2_angle - n1.second) > 180)
		if (n2_angle < n1.second) n2_angle += 360; else n2_angle -= 360;

	// n1, n2 now are the surrounding two norm nodes.
	double alpha = (l - n1.first) / (n2.first - n1.first);

	final_dir.normalize();
	Point p = Point(final_pos, triple(0, 0, 1), final_dir);

	triple dir, out, up;
	dir = p.direction();
	up = p.up_out().first;
	out = dir.rotate(up, 90);

	double theta = alpha * n2_angle + (1 - alpha) * n1.second;

	triple final_up = up.rotate(dir, theta);
		
	return Point(final_pos, final_up, final_dir);
}





std::vector<Point> Curve::curve_points() {
	switch (Config::instance().spline()) {
		case B_SPLINE:
			points_ = b_spline();
			break;
		case CARDINAL:
			points_ = cardinal();
			break;
		default:
			points_ = b_spline();
	}
	
	vector<Point> points;
	for (double length = 1; length < length_.back(); length += 3)
		points.push_back(point_at_length(length));

	return points;	
}

vector<Point> Curve::b_spline() {
	double mult = 1.0 / 6.0;
	double mat[16] = {
			1, 4, 1, 0,
			-3, 0, 3, 0,
			3, -6, 3, 0,
			-1, 3, -3, 1
			};
	return make_curve(mult, mat);
}

vector<Point> Curve::make_curve(double mult, double mat[16]) {
	std::vector<Point> points;

	vector<double> cp[4];
	double a[3][4];
	double s = 0.5;

	length_.clear();
	length_.push_back(0);

	for (int index = 0; index < controls.size(); index++)
	{
		cp[0] = controls[index].position();
		cp[1] = controls[(index + 1) % controls.size()].position();
		cp[2] = controls[(index + 2) % controls.size()].position();
		cp[3] = controls[(index + 3) % controls.size()].position();

		for (int i = 0; i < 3; i++) {
			for (int j = 0; j < 4; j++)
			{
				a[i][j] = 0;
				for (int k = 0; k < 4; k++)
					a[i][j] += mat[4 * j + k] * cp[k][i];
				a[i][j] *= mult;
			}

		}

		double x, y, z, next_x, next_y, next_z, t, adjust;

		if (index == 0) {
			x = a[0][0];
			y = a[1][0];
			z = a[2][0];
			t = 0.1;
		}
		

		for (; t<= 1; t+=0.05)
		{
			next_x = a[0][3];
			next_y = a[1][3];
			next_z = a[2][3];
			for (int i = 2; i >= 0; --i) {
				next_x = next_x * t + a[0][i];
				next_y = next_y * t + a[1][i];
				next_z = next_z * t + a[2][i];
			}
			
			points.push_back(Point(x, y, z, x, y, z, next_x - x, next_y - y, next_z - z));

			Point::triple dL = points.back().direction();
			length_.push_back(length_.back() + magnitude(dL.x, dL.y, dL.z));

			x = next_x;
			y = next_y;
			z = next_z;
		}

		t = t - 1;
	}

	length_.pop_back();

	double dx, dy, dz;
	Point::triple p1, p2;
	p1 = points.begin()->position();
	p2 = (points.end() - 1)->position();

	length_.push_back(length_.back() + magnitude(p1.x - p2.x, p1.y - p2.y, p1.z - p2.z));

	points.push_back(points[0]);

	points_ = points;

	if (normals.empty()) {
		normals.push_back(pair<double, double>(0, 0));
	}
	while (normals.back().first + 100 < length_.back())
	{
		normals.push_back(pair<double, double>(normals.back().first + 100, 0));
	}

	while (normals.back().first > length_.back())
	{
		normals.pop_back();
	}
	

	return points;
}



std::vector<Point> Curve::cardinal() {
	double s = 0.5;
	double mat[16] = {
		    0,     1,         0,  0,
		   -s,     0,         s,  0,
		2 * s, s - 3, 3 - 2 * s, -s,
		   -s, 2 - s,     s - 2,  s
	};
	

	return make_curve(1, mat);
}

double Curve::magnitude(double x, double y, double z)
{
	return sqrt(x * x + y * y + z * z);
}